@@ -28,6 +28,7 @@ import six
 from stevedore import driver
 
 import tooz
+from tooz import _retry
 
 LOG = logging.getLogger(__name__)
 
@@ -126,6 +127,11 @@ class MemberJoinedGroup(Event):
         self.group_id = group_id
         self.member_id = member_id
 
+    def __repr__(self):
+        return "<%s: group %s: +member %s>" % (self.__class__.__name__,
+                                               self.member_id,
+                                               self.group_id)
+
 
 class MemberLeftGroup(Event):
     """A member left a group event."""
@@ -134,6 +140,11 @@ class MemberLeftGroup(Event):
         self.group_id = group_id
         self.member_id = member_id
 
+    def __repr__(self):
+        return "<%s: group %s: -member %s>" % (self.__class__.__name__,
+                                               self.member_id,
+                                               self.group_id)
+
 
 class LeaderElected(Event):
     """A leader as been elected."""
@@ -214,7 +225,6 @@ class Heart(object):
         return self._finished.is_set()
 
 
-@six.add_metaclass(abc.ABCMeta)
 class CoordinationDriver(object):
 
     requires_beating = False
@@ -231,21 +241,21 @@ class CoordinationDriver(object):
     enum member(s) that can be used to interogate how this driver works.
     """
 
-    def __init__(self):
+    def __init__(self, member_id):
+        super(CoordinationDriver, self).__init__()
+        self._member_id = member_id
         self._started = False
         self._hooks_join_group = collections.defaultdict(Hooks)
         self._hooks_leave_group = collections.defaultdict(Hooks)
         self._hooks_elected_leader = collections.defaultdict(Hooks)
-        # A cache for group members
-        self._group_members = collections.defaultdict(set)
         self.requires_beating = (
             CoordinationDriver.heartbeat != self.__class__.heartbeat
         )
         self.heart = Heart(self)
 
     def _has_hooks_for_group(self, group_id):
-        return (len(self._hooks_join_group[group_id]) +
-                len(self._hooks_leave_group[group_id]))
+        return (group_id in self._hooks_join_group or
+                group_id in self._hooks_leave_group)
 
     @staticmethod
     def run_watchers(timeout=None):
@@ -261,7 +271,6 @@ class CoordinationDriver(object):
         """Try to leader elect this coordinator & activate hooks on success."""
         raise tooz.NotImplemented
 
-    @abc.abstractmethod
     def watch_join_group(self, group_id, callback):
         """Call a function when group_id sees a new member joined.
 
@@ -274,7 +283,6 @@ class CoordinationDriver(object):
         """
         self._hooks_join_group[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_join_group(self, group_id, callback):
         """Stop executing a function when a group_id sees a new member joined.
 
@@ -283,15 +291,17 @@ class CoordinationDriver(object):
                          this group
         """
         try:
+            # Check if group_id is in hooks to avoid creating a default empty
+            # entry in hooks list.
+            if group_id not in self._hooks_join_group:
+                raise ValueError
             self._hooks_join_group[group_id].remove(callback)
         except ValueError:
             raise WatchCallbackNotFound(group_id, callback)
 
-        if (not self._has_hooks_for_group(group_id) and
-           group_id in self._group_members):
-            del self._group_members[group_id]
+        if not self._hooks_join_group[group_id]:
+            del self._hooks_join_group[group_id]
 
-    @abc.abstractmethod
     def watch_leave_group(self, group_id, callback):
         """Call a function when group_id sees a new member leaving.
 
@@ -305,7 +315,6 @@ class CoordinationDriver(object):
         """
         self._hooks_leave_group[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_leave_group(self, group_id, callback):
         """Stop executing a function when a group_id sees a new member leaving.
 
@@ -314,18 +323,17 @@ class CoordinationDriver(object):
                          this group
         """
         try:
+            # Check if group_id is in hooks to avoid creating a default empty
+            # entry in hooks list.
+            if group_id not in self._hooks_leave_group:
+                raise ValueError
             self._hooks_leave_group[group_id].remove(callback)
         except ValueError:
             raise WatchCallbackNotFound(group_id, callback)
 
-        if (not self._has_hooks_for_group(group_id) and
-           group_id in self._group_members):
-            del self._group_members[group_id]
-
         if not self._hooks_leave_group[group_id]:
             del self._hooks_leave_group[group_id]
 
-    @abc.abstractmethod
     def watch_elected_as_leader(self, group_id, callback):
         """Call a function when member gets elected as leader.
 
@@ -339,7 +347,6 @@ class CoordinationDriver(object):
         """
         self._hooks_elected_leader[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_elected_as_leader(self, group_id, callback):
         """Call a function when member gets elected as leader.
 
@@ -384,6 +391,8 @@ class CoordinationDriver(object):
         if self.requires_beating and start_heart:
             self.heart.start()
         self._started = True
+        # Tracks which group are joined
+        self._joined_groups = set()
 
     def _start(self):
         pass
@@ -399,6 +408,17 @@ class CoordinationDriver(object):
         if self.heart.is_alive():
             self.heart.stop()
             self.heart.wait()
+        leaving = [self.leave_group(group)
+                   for group in self._joined_groups]
+        for leave in leaving:
+            try:
+                leave.get()
+            except ToozError:
+                # Whatever happens, ignore. Maybe we got booted out/never
+                # existed in the first place, or something is down, but we just
+                # want to call _stop after whatever happens to not leak any
+                # connection.
+                pass
         self._stop()
         self._started = False
 
@@ -438,6 +458,33 @@ class CoordinationDriver(object):
         """
         raise tooz.NotImplemented
 
+    @_retry.retry()
+    def join_group_create(self, group_id):
+        """Join a group and create it if necessary.
+
+        If the group cannot be joined because it does not exist, it is created
+        before being joined.
+
+        This function will keep retrying until it can create the group and join
+        it. Since nothing is transactional, it may have to retry several times
+        if another member is creating/deleting the group at the same time.
+
+        :param group_id: Identifier of the group to join and create
+
+        """
+        req = self.join_group(group_id)
+        try:
+            req.get()
+        except GroupNotCreated:
+            req = self.create_group(group_id)
+            try:
+                req.get()
+            except GroupAlreadyExist:
+                # The group might have been created in the meantime, ignore
+                pass
+            # Now retry to join the group
+            raise _retry.TryAgain
+
     @staticmethod
     def leave_group(group_id):
         """Leave a group asynchronously.
@@ -462,9 +509,9 @@ class CoordinationDriver(object):
 
     @staticmethod
     def get_members(group_id):
-        """Return the list of all members ids of the specified group.
+        """Return the set of all members ids of the specified group.
 
-        :returns: list of all created group ids
+        :returns: set of all created group ids
         :rtype: CoordAsyncResult
         """
         raise tooz.NotImplemented
@@ -565,23 +612,63 @@ class CoordAsyncResult(object):
         """Returns True if the task is done, False otherwise."""
 
 
-class _RunWatchersMixin(object):
-    """Mixin to share the *mostly* common ``run_watchers`` implementation."""
+class CoordinationDriverCachedRunWatchers(CoordinationDriver):
+    """Coordination driver with a `run_watchers` implementation.
+
+    This implementation of `run_watchers` is based on a cache of the group
+    members between each run of `run_watchers` that is being updated between
+    each run.
+
+    """
+
+    def __init__(self, member_id):
+        super(CoordinationDriverCachedRunWatchers, self).__init__(member_id)
+        # A cache for group members
+        self._group_members = collections.defaultdict(set)
+        self._joined_groups = set()
+
+    def _init_watch_group(self, group_id):
+        members = self.get_members(group_id).get()
+        if group_id not in self._group_members:
+            self._group_members[group_id] = members
+
+    def watch_join_group(self, group_id, callback):
+        self._init_watch_group(group_id)
+        super(CoordinationDriverCachedRunWatchers, self).watch_join_group(
+            group_id, callback)
+
+    def unwatch_join_group(self, group_id, callback):
+        super(CoordinationDriverCachedRunWatchers, self).unwatch_join_group(
+            group_id, callback)
+
+        if (not self._has_hooks_for_group(group_id) and
+           group_id in self._group_members):
+            del self._group_members[group_id]
+
+    def watch_leave_group(self, group_id, callback):
+        self._init_watch_group(group_id)
+        super(CoordinationDriverCachedRunWatchers, self).watch_leave_group(
+            group_id, callback)
+
+    def unwatch_leave_group(self, group_id, callback):
+        super(CoordinationDriverCachedRunWatchers, self).unwatch_leave_group(
+            group_id, callback)
+
+        if (not self._has_hooks_for_group(group_id) and
+           group_id in self._group_members):
+            del self._group_members[group_id]
 
     def run_watchers(self, timeout=None):
         with timeutils.StopWatch(duration=timeout) as w:
-            known_groups = self.get_groups().get(
-                timeout=w.leftover(return_none=True))
             result = []
-            for group_id in known_groups:
+            group_with_hooks = set(self._hooks_join_group.keys()).union(
+                set(self._hooks_leave_group.keys()))
+            for group_id in group_with_hooks:
                 try:
-                    group_members_fut = self.get_members(group_id)
-                    group_members = group_members_fut.get(
+                    group_members = self.get_members(group_id).get(
                         timeout=w.leftover(return_none=True))
                 except GroupNotCreated:
                     group_members = set()
-                else:
-                    group_members = set(group_members)
                 if (group_id in self._joined_groups and
                         self._member_id not in group_members):
                     self._joined_groups.discard(group_id)
